/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#ifndef COMMON_LAPACK
#define COMMON_LAPACK

#ifndef ASSEMBLER

/* Lapack Library */

blasint sgetf2_k(long, long, float   *a, long, blasint *, long, float   *);
blasint dgetf2_k(long, long, double  *a, long, blasint *, long, double  *);
blasint qgetf2_k(long, long, xdouble *a, long, blasint *, long, xdouble *);
blasint cgetf2_k(long, long, float   *a, long, blasint *, long, float   *);
blasint zgetf2_k(long, long, double  *a, long, blasint *, long, double  *);
blasint xgetf2_k(long, long, xdouble *a, long, blasint *, long, xdouble *);

blasint sgetrf_single(long, long, float   *a, long, blasint *, long, float   *, float   *);
blasint dgetrf_single(long, long, double  *a, long, blasint *, long, double  *, double  *);
blasint qgetrf_single(long, long, xdouble *a, long, blasint *, long, xdouble *, xdouble *);
blasint cgetrf_single(long, long, float   *a, long, blasint *, long, float   *, float   *);
blasint zgetrf_single(long, long, double  *a, long, blasint *, long, double  *, double  *);
blasint xgetrf_single(long, long, xdouble *a, long, blasint *, long, xdouble *, xdouble *);

blasint sgetrf_parallel(long, long, float   *a, long, blasint *, long, float   *);
blasint dgetrf_parallel(long, long, double  *a, long, blasint *, long, double  *);
blasint qgetrf_parallel(long, long, xdouble *a, long, blasint *, long, xdouble *);
blasint cgetrf_parallel(long, long, float   *a, long, blasint *, long, float   *);
blasint zgetrf_parallel(long, long, double  *a, long, blasint *, long, double  *);
blasint xgetrf_parallel(long, long, xdouble *a, long, blasint *, long, xdouble *);

int slaswp_plus (long, long, long, float,   float   *, long, float  *, long, blasint *, long);
int slaswp_minus(long, long, long, float,   float   *, long, float  *, long, blasint *, long);
int dlaswp_plus (long, long, long, double,  double  *, long, double *, long, blasint *, long);
int dlaswp_minus(long, long, long, double,  double  *, long, double *, long, blasint *, long);
int qlaswp_plus (long, long, long, xdouble, xdouble *, long, xdouble *, long, blasint *, long);
int qlaswp_minus(long, long, long, xdouble, xdouble *, long, xdouble *, long, blasint *, long);

int claswp_plus (long, long, long, float,   float,   float   *, long, float   *, long, blasint *, long);
int claswp_minus(long, long, long, float,   float,   float   *, long, float   *, long, blasint *, long);
int zlaswp_plus (long, long, long, double,  double,  double  *, long, double  *, long, blasint *, long);
int zlaswp_minus(long, long, long, double,  double,  double  *, long, double  *, long, blasint *, long);
int xlaswp_plus (long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, blasint *, long);
int xlaswp_minus(long, long, long, xdouble, xdouble, xdouble *, long, xdouble *, long, blasint *, long);

int slaswp_ncopy(long, long, long, float   *, long, blasint *, float   *);
int dlaswp_ncopy(long, long, long, double  *, long, blasint *, double  *);
int qlaswp_ncopy(long, long, long, xdouble *, long, blasint *, xdouble *);
int claswp_ncopy(long, long, long, float   *, long, blasint *, float   *);
int zlaswp_ncopy(long, long, long, double  *, long, blasint *, double  *);
int xlaswp_ncopy(long, long, long, xdouble *, long, blasint *, xdouble *);

blasint sgetrs_N_single(long, long, float   *, long, float   *, long, blasint *, float   *, float   *);
blasint sgetrs_T_single(long, long, float   *, long, float   *, long, blasint *, float   *, float   *);
blasint dgetrs_N_single(long, long, double  *, long, double  *, long, blasint *, double  *, double  *);
blasint dgetrs_T_single(long, long, double  *, long, double  *, long, blasint *, double  *, double  *);
blasint qgetrs_N_single(long, long, xdouble *, long, xdouble *, long, blasint *, xdouble *, xdouble *);
blasint qgetrs_T_single(long, long, xdouble *, long, xdouble *, long, blasint *, xdouble *, xdouble *);
blasint cgetrs_N_single(long, long, float   *, long, float   *, long, blasint *, float   *, float   *);
blasint cgetrs_T_single(long, long, float   *, long, float   *, long, blasint *, float   *, float   *);
blasint cgetrs_R_single(long, long, float   *, long, float   *, long, blasint *, float   *, float   *);
blasint cgetrs_C_single(long, long, float   *, long, float   *, long, blasint *, float   *, float   *);
blasint zgetrs_N_single(long, long, double  *, long, double  *, long, blasint *, double  *, double  *);
blasint zgetrs_T_single(long, long, double  *, long, double  *, long, blasint *, double  *, double  *);
blasint zgetrs_R_single(long, long, double  *, long, double  *, long, blasint *, double  *, double  *);
blasint zgetrs_C_single(long, long, double  *, long, double  *, long, blasint *, double  *, double  *);
blasint xgetrs_N_single(long, long, xdouble *, long, xdouble *, long, blasint *, xdouble *, xdouble *);
blasint xgetrs_T_single(long, long, xdouble *, long, xdouble *, long, blasint *, xdouble *, xdouble *);
blasint xgetrs_R_single(long, long, xdouble *, long, xdouble *, long, blasint *, xdouble *, xdouble *);
blasint xgetrs_C_single(long, long, xdouble *, long, xdouble *, long, blasint *, xdouble *, xdouble *);

blasint sgetrs_N_parallel(long, long, float   *, long, float   *, long, blasint *, float   *);
blasint sgetrs_T_parallel(long, long, float   *, long, float   *, long, blasint *, float   *);
blasint dgetrs_N_parallel(long, long, double  *, long, double  *, long, blasint *, double  *);
blasint dgetrs_T_parallel(long, long, double  *, long, double  *, long, blasint *, double  *);
blasint qgetrs_N_parallel(long, long, xdouble *, long, xdouble *, long, blasint *, xdouble *);
blasint qgetrs_T_parallel(long, long, xdouble *, long, xdouble *, long, blasint *, xdouble *);
blasint cgetrs_N_parallel(long, long, float   *, long, float   *, long, blasint *, float   *);
blasint cgetrs_T_parallel(long, long, float   *, long, float   *, long, blasint *, float   *);
blasint cgetrs_R_parallel(long, long, float   *, long, float   *, long, blasint *, float   *);
blasint cgetrs_C_parallel(long, long, float   *, long, float   *, long, blasint *, float   *);
blasint zgetrs_N_parallel(long, long, double  *, long, double  *, long, blasint *, double  *);
blasint zgetrs_T_parallel(long, long, double  *, long, double  *, long, blasint *, double  *);
blasint zgetrs_R_parallel(long, long, double  *, long, double  *, long, blasint *, double  *);
blasint zgetrs_C_parallel(long, long, double  *, long, double  *, long, blasint *, double  *);
blasint xgetrs_N_parallel(long, long, xdouble *, long, xdouble *, long, blasint *, xdouble *);
blasint xgetrs_T_parallel(long, long, xdouble *, long, xdouble *, long, blasint *, xdouble *);
blasint xgetrs_R_parallel(long, long, xdouble *, long, xdouble *, long, blasint *, xdouble *);
blasint xgetrs_C_parallel(long, long, xdouble *, long, xdouble *, long, blasint *, xdouble *);

blasint spotf2_U(long, float   *, long, float   *);
blasint spotf2_L(long, float   *, long, float   *);
blasint dpotf2_U(long, double  *, long, double  *);
blasint dpotf2_L(long, double  *, long, double  *);
blasint qpotf2_U(long, xdouble *, long, xdouble *);
blasint qpotf2_L(long, xdouble *, long, xdouble *);
blasint cpotf2_U(long, float   *, long, float   *);
blasint cpotf2_L(long, float   *, long, float   *);
blasint zpotf2_U(long, double  *, long, double  *);
blasint zpotf2_L(long, double  *, long, double  *);
blasint xpotf2_U(long, xdouble *, long, xdouble *);
blasint xpotf2_L(long, xdouble *, long, xdouble *);

blasint spotrf_U_single(long, float   *, long, float   *, float   *);
blasint spotrf_L_single(long, float   *, long, float   *, float   *);
blasint dpotrf_U_single(long, double  *, long, double  *, double  *);
blasint dpotrf_L_single(long, double  *, long, double  *, double  *);
blasint qpotrf_U_single(long, xdouble *, long, xdouble *, xdouble *);
blasint qpotrf_L_single(long, xdouble *, long, xdouble *, xdouble *);
blasint cpotrf_U_single(long, float   *, long, float   *, float   *);
blasint cpotrf_L_single(long, float   *, long, float   *, float   *);
blasint zpotrf_U_single(long, double  *, long, double  *, double  *);
blasint zpotrf_L_single(long, double  *, long, double  *, double  *);
blasint xpotrf_U_single(long, xdouble *, long, xdouble *, xdouble *);
blasint xpotrf_L_single(long, xdouble *, long, xdouble *, xdouble *);

blasint spotrf_U_parallel(long, float   *, long, float   *);
blasint spotrf_L_parallel(long, float   *, long, float   *);
blasint dpotrf_U_parallel(long, double  *, long, double  *);
blasint dpotrf_L_parallel(long, double  *, long, double  *);
blasint qpotrf_U_parallel(long, xdouble *, long, xdouble *);
blasint qpotrf_L_parallel(long, xdouble *, long, xdouble *);
blasint cpotrf_U_parallel(long, float   *, long, float   *);
blasint cpotrf_L_parallel(long, float   *, long, float   *);
blasint zpotrf_U_parallel(long, double  *, long, double  *);
blasint zpotrf_L_parallel(long, double  *, long, double  *);
blasint xpotrf_U_parallel(long, xdouble *, long, xdouble *);
blasint xpotrf_L_parallel(long, xdouble *, long, xdouble *);

blasint slauu2_U(long, float   *, long, float   *);
blasint slauu2_L(long, float   *, long, float   *);
blasint dlauu2_U(long, double  *, long, double  *);
blasint dlauu2_L(long, double  *, long, double  *);
blasint qlauu2_U(long, xdouble *, long, xdouble *);
blasint qlauu2_L(long, xdouble *, long, xdouble *);
blasint clauu2_U(long, float   *, long, float   *);
blasint clauu2_L(long, float   *, long, float   *);
blasint zlauu2_U(long, double  *, long, double  *);
blasint zlauu2_L(long, double  *, long, double  *);
blasint xlauu2_U(long, xdouble *, long, xdouble *);
blasint xlauu2_L(long, xdouble *, long, xdouble *);

blasint slauum_U_single(long, float   *, long, float   *, float *);
blasint slauum_L_single(long, float   *, long, float   *, float *);
blasint qlauum_L_single(long, xdouble *, long, xdouble *, xdouble *);
blasint qlauum_U_single(long, xdouble *, long, xdouble *, xdouble *);
blasint dlauum_U_single(long, double  *, long, double  *, double *);
blasint dlauum_L_single(long, double  *, long, double  *, double *);
blasint clauum_U_single(long, float   *, long, float   *, float *);
blasint clauum_L_single(long, float   *, long, float   *, float *);
blasint zlauum_U_single(long, double  *, long, double  *, double *);
blasint zlauum_L_single(long, double  *, long, double  *, double *);
blasint xlauum_U_single(long, xdouble *, long, xdouble *, xdouble *);
blasint xlauum_L_single(long, xdouble *, long, xdouble *, xdouble *);

blasint slauum_U_parallel(long, float   *, long, float   *);
blasint slauum_L_parallel(long, float   *, long, float   *);
blasint dlauum_U_parallel(long, double  *, long, double  *);
blasint dlauum_L_parallel(long, double  *, long, double  *);
blasint qlauum_L_parallel(long, xdouble *, long, xdouble *);
blasint qlauum_U_parallel(long, xdouble *, long, xdouble *);
blasint clauum_U_parallel(long, float   *, long, float   *);
blasint clauum_L_parallel(long, float   *, long, float   *);
blasint zlauum_U_parallel(long, double  *, long, double  *);
blasint zlauum_L_parallel(long, double  *, long, double  *);
blasint xlauum_U_parallel(long, xdouble *, long, xdouble *);
blasint xlauum_L_parallel(long, xdouble *, long, xdouble *);

blasint strti2_UU(long, float   *, long, float   *);
blasint strti2_UN(long, float   *, long, float   *);
blasint strti2_LU(long, float   *, long, float   *);
blasint strti2_LN(long, float   *, long, float   *);
blasint dtrti2_UU(long, double  *, long, double  *);
blasint dtrti2_UN(long, double  *, long, double  *);
blasint dtrti2_LU(long, double  *, long, double  *);
blasint dtrti2_LN(long, double  *, long, double  *);
blasint qtrti2_UU(long, xdouble *, long, xdouble *);
blasint qtrti2_UN(long, xdouble *, long, xdouble *);
blasint qtrti2_LU(long, xdouble *, long, xdouble *);
blasint qtrti2_LN(long, xdouble *, long, xdouble *);
blasint ctrti2_UU(long, float   *, long, float   *);
blasint ctrti2_UN(long, float   *, long, float   *);
blasint ctrti2_LU(long, float   *, long, float   *);
blasint ctrti2_LN(long, float   *, long, float   *);
blasint ztrti2_UU(long, double  *, long, double  *);
blasint ztrti2_UN(long, double  *, long, double  *);
blasint ztrti2_LU(long, double  *, long, double  *);
blasint ztrti2_LN(long, double  *, long, double  *);
blasint xtrti2_UU(long, xdouble *, long, xdouble *);
blasint xtrti2_UN(long, xdouble *, long, xdouble *);
blasint xtrti2_LU(long, xdouble *, long, xdouble *);
blasint xtrti2_LN(long, xdouble *, long, xdouble *);

blasint strtri_UU_single(long, float   *, long, float   *, float   *);
blasint strtri_UN_single(long, float   *, long, float   *, float   *);
blasint strtri_LU_single(long, float   *, long, float   *, float   *);
blasint strtri_LN_single(long, float   *, long, float   *, float   *);
blasint dtrtri_UU_single(long, double  *, long, double  *, double  *);
blasint dtrtri_UN_single(long, double  *, long, double  *, double  *);
blasint dtrtri_LU_single(long, double  *, long, double  *, double  *);
blasint dtrtri_LN_single(long, double  *, long, double  *, double  *);
blasint qtrtri_UU_single(long, xdouble *, long, xdouble *, xdouble *);
blasint qtrtri_UN_single(long, xdouble *, long, xdouble *, xdouble *);
blasint qtrtri_LU_single(long, xdouble *, long, xdouble *, xdouble *);
blasint qtrtri_LN_single(long, xdouble *, long, xdouble *, xdouble *);
blasint ctrtri_UU_single(long, float   *, long, float   *, float   *);
blasint ctrtri_UN_single(long, float   *, long, float   *, float   *);
blasint ctrtri_LU_single(long, float   *, long, float   *, float   *);
blasint ctrtri_LN_single(long, float   *, long, float   *, float   *);
blasint ztrtri_UU_single(long, double  *, long, double  *, double  *);
blasint ztrtri_UN_single(long, double  *, long, double  *, double  *);
blasint ztrtri_LU_single(long, double  *, long, double  *, double  *);
blasint ztrtri_LN_single(long, double  *, long, double  *, double  *);
blasint xtrtri_UU_single(long, xdouble *, long, xdouble *, xdouble *);
blasint xtrtri_UN_single(long, xdouble *, long, xdouble *, xdouble *);
blasint xtrtri_LU_single(long, xdouble *, long, xdouble *, xdouble *);
blasint xtrtri_LN_single(long, xdouble *, long, xdouble *, xdouble *);

blasint strtri_UU_parallel(long, float   *, long, float   *);
blasint strtri_UN_parallel(long, float   *, long, float   *);
blasint strtri_LU_parallel(long, float   *, long, float   *);
blasint strtri_LN_parallel(long, float   *, long, float   *);
blasint dtrtri_UU_parallel(long, double  *, long, double  *);
blasint dtrtri_UN_parallel(long, double  *, long, double  *);
blasint dtrtri_LU_parallel(long, double  *, long, double  *);
blasint dtrtri_LN_parallel(long, double  *, long, double  *);
blasint qtrtri_UU_parallel(long, xdouble *, long, xdouble *);
blasint qtrtri_UN_parallel(long, xdouble *, long, xdouble *);
blasint qtrtri_LU_parallel(long, xdouble *, long, xdouble *);
blasint qtrtri_LN_parallel(long, xdouble *, long, xdouble *);
blasint ctrtri_UU_parallel(long, float   *, long, float   *);
blasint ctrtri_UN_parallel(long, float   *, long, float   *);
blasint ctrtri_LU_parallel(long, float   *, long, float   *);
blasint ctrtri_LN_parallel(long, float   *, long, float   *);
blasint ztrtri_UU_parallel(long, double  *, long, double  *);
blasint ztrtri_UN_parallel(long, double  *, long, double  *);
blasint ztrtri_LU_parallel(long, double  *, long, double  *);
blasint ztrtri_LN_parallel(long, double  *, long, double  *);
blasint xtrtri_UU_parallel(long, xdouble *, long, xdouble *);
blasint xtrtri_UN_parallel(long, xdouble *, long, xdouble *);
blasint xtrtri_LU_parallel(long, xdouble *, long, xdouble *);
blasint xtrtri_LN_parallel(long, xdouble *, long, xdouble *);

void sneg_tcopy(long, long, float   *, long, float   *);
void dneg_tcopy(long, long, double  *, long, double  *);
void qneg_tcopy(long, long, xdouble *, long, xdouble *);
void cneg_tcopy(long, long, float   *, long, float   *);
void zneg_tcopy(long, long, double  *, long, double  *);
void xneg_tcopy(long, long, xdouble *, long, xdouble *);

#define SGETF2	sgetf2_k
#define DGETF2	dgetf2_k
#define QGETF2	qgetf2_k
#define CGETF2	cgetf2_k
#define ZGETF2	zgetf2_k
#define XGETF2	xgetf2_k

#define SLASWP_PLUS	slaswp_plus
#define SLASWP_MINUS	slaswp_minus
#define DLASWP_PLUS	dlaswp_plus
#define DLASWP_MINUS	dlaswp_minus
#define QLASWP_PLUS	qlaswp_plus
#define QLASWP_MINUS	qlaswp_minus
#define CLASWP_PLUS	claswp_plus
#define CLASWP_MINUS	claswp_minus
#define ZLASWP_PLUS	zlaswp_plus
#define ZLASWP_MINUS	zlaswp_minus
#define XLASWP_PLUS	xlaswp_plus
#define XLASWP_MINUS	xlaswp_minus

#define SLASWP_NCOPY	slaswp_ncopy
#define DLASWP_NCOPY	dlaswp_ncopy
#define QLASWP_NCOPY	qlaswp_ncopy
#define CLASWP_NCOPY	claswp_ncopy
#define ZLASWP_NCOPY	zlaswp_ncopy
#define XLASWP_NCOPY	xlaswp_ncopy

#define SNEG_TCOPY	sneg_tcopy
#define DNEG_TCOPY	dneg_tcopy
#define QNEG_TCOPY	qneg_tcopy
#define CNEG_TCOPY	cneg_tcopy
#define ZNEG_TCOPY	zneg_tcopy
#define XNEG_TCOPY	xneg_tcopy

#ifndef COMPLEX
#ifdef XDOUBLE
#define GETF2	QGETF2
#define GETRF	QGETRF
#define GETRS_N_SINGLE	qgetrs_N_single
#define GETRS_T_SINGLE	qgetrs_T_single
#define GETRS_R_SINGLE	qgetrs_N_single
#define GETRS_C_SINGLE	qgetrs_T_single
#define GETRS_N_PARALLEL	qgetrs_N_parallel
#define GETRS_T_PARALLEL	qgetrs_T_parallel
#define GETRS_R_PARALLEL	qgetrs_N_parallel
#define GETRS_C_PARALLEL	qgetrs_T_parallel
#define LASWP_PLUS	QLASWP_PLUS
#define LASWP_MINUS	QLASWP_MINUS
#define LASWP_NCOPY	QLASWP_NCOPY
#define GETRS_N		QGETRS_N
#define GETRS_T		QGETRS_T
#define GETRF_SINGLE	qgetrf_single
#define GETRF_PARALLEL	qgetrf_parallel
#define NEG_TCOPY	QNEG_TCOPY
#elif defined(DOUBLE)
#define GETF2	DGETF2
#define GETRF	DGETRF
#define GETRS_N_SINGLE	dgetrs_N_single
#define GETRS_T_SINGLE	dgetrs_T_single
#define GETRS_R_SINGLE	dgetrs_N_single
#define GETRS_C_SINGLE	dgetrs_T_single
#define GETRS_N_PARALLEL	dgetrs_N_parallel
#define GETRS_T_PARALLEL	dgetrs_T_parallel
#define GETRS_R_PARALLEL	dgetrs_N_parallel
#define GETRS_C_PARALLEL	dgetrs_T_parallel
#define LASWP_PLUS	DLASWP_PLUS
#define LASWP_MINUS	DLASWP_MINUS
#define LASWP_NCOPY	DLASWP_NCOPY
#define GETRS_N		DGETRS_N
#define GETRS_T		DGETRS_T
#define GETRF_SINGLE	dgetrf_single
#define GETRF_PARALLEL	dgetrf_parallel
#define NEG_TCOPY	DNEG_TCOPY
#else
#define GETF2	SGETF2
#define GETRF	SGETRF
#define GETRS_N_SINGLE	sgetrs_N_single
#define GETRS_T_SINGLE	sgetrs_T_single
#define GETRS_R_SINGLE	sgetrs_N_single
#define GETRS_C_SINGLE	sgetrs_T_single
#define GETRS_N_PARALLEL	sgetrs_N_parallel
#define GETRS_T_PARALLEL	sgetrs_T_parallel
#define GETRS_R_PARALLEL	sgetrs_N_parallel
#define GETRS_C_PARALLEL	sgetrs_T_parallel
#define LASWP_PLUS	SLASWP_PLUS
#define LASWP_MINUS	SLASWP_MINUS
#define LASWP_NCOPY	SLASWP_NCOPY
#define GETRS_N		SGETRS_N
#define GETRS_T		SGETRS_T
#define GETRF_SINGLE	sgetrf_single
#define GETRF_PARALLEL	sgetrf_parallel
#define NEG_TCOPY	SNEG_TCOPY
#endif
#else
#ifdef XDOUBLE
#define GETF2	XGETF2
#define GETRF	XGETRF
#define GETRS_N_SINGLE	xgetrs_N_single
#define GETRS_T_SINGLE	xgetrs_T_single
#define GETRS_R_SINGLE	xgetrs_R_single
#define GETRS_C_SINGLE	xgetrs_C_single
#define GETRS_N_PARALLEL	xgetrs_N_parallel
#define GETRS_T_PARALLEL	xgetrs_T_parallel
#define GETRS_R_PARALLEL	xgetrs_R_parallel
#define GETRS_C_PARALLEL	xgetrs_C_parallel
#define LASWP_PLUS	XLASWP_PLUS
#define LASWP_MINUS	XLASWP_MINUS
#define LASWP_NCOPY	XLASWP_NCOPY
#define GETRS_N		XGETRS_N
#define GETRS_T		XGETRS_T
#define GETRF_SINGLE	xgetrf_single
#define GETRF_PARALLEL	xgetrf_parallel
#define NEG_TCOPY	XNEG_TCOPY
#elif defined(DOUBLE)
#define GETF2	ZGETF2
#define GETRF	ZGETRF
#define GETRS_N_SINGLE	zgetrs_N_single
#define GETRS_T_SINGLE	zgetrs_T_single
#define GETRS_R_SINGLE	zgetrs_R_single
#define GETRS_C_SINGLE	zgetrs_C_single
#define GETRS_N_PARALLEL	zgetrs_N_parallel
#define GETRS_T_PARALLEL	zgetrs_T_parallel
#define GETRS_R_PARALLEL	zgetrs_R_parallel
#define GETRS_C_PARALLEL	zgetrs_C_parallel
#define LASWP_PLUS	ZLASWP_PLUS
#define LASWP_MINUS	ZLASWP_MINUS
#define LASWP_NCOPY	ZLASWP_NCOPY
#define GETRS_N		ZGETRS_N
#define GETRS_T		ZGETRS_T
#define GETRF_SINGLE	zgetrf_single
#define GETRF_PARALLEL	zgetrf_parallel
#define NEG_TCOPY	ZNEG_TCOPY
#else
#define GETF2	CGETF2
#define GETRF	CGETRF
#define GETRS_N_SINGLE	cgetrs_N_single
#define GETRS_T_SINGLE	cgetrs_T_single
#define GETRS_R_SINGLE	cgetrs_R_single
#define GETRS_C_SINGLE	cgetrs_C_single
#define GETRS_N_PARALLEL	cgetrs_N_parallel
#define GETRS_T_PARALLEL	cgetrs_T_parallel
#define GETRS_R_PARALLEL	cgetrs_R_parallel
#define GETRS_C_PARALLEL	cgetrs_C_parallel
#define LASWP_PLUS	CLASWP_PLUS
#define LASWP_MINUS	CLASWP_MINUS
#define LASWP_NCOPY	CLASWP_NCOPY
#define GETRS_N		CGETRS_N
#define GETRS_T		CGETRS_T
#define GETRF_SINGLE	cgetrf_single
#define GETRF_PARALLEL	cgetrf_parallel
#define NEG_TCOPY	CNEG_TCOPY
#endif
#endif

#ifndef COMPLEX
#ifdef XDOUBLE
#define  POTF2_U  qpotf2_U
#define  POTF2_L  qpotf2_L
#define  LAUU2_U  qlauu2_U
#define  LAUU2_L  qlauu2_L
#define  POTRF_U_SINGLE qpotrf_U_single
#define  POTRF_L_SINGLE qpotrf_L_single
#define  POTRF_U_PARALLEL qpotrf_U_parallel
#define  POTRF_L_PARALLEL qpotrf_L_parallel
#define  LAUUM_U_SINGLE qlauum_U_single
#define  LAUUM_L_SINGLE qlauum_L_single
#define  LAUUM_U_PARALLEL qlauum_U_parallel
#define  LAUUM_L_PARALLEL qlauum_L_parallel
#define  TRTI2_UU qtrti2_UU
#define  TRTI2_UN qtrti2_UN
#define  TRTI2_LU qtrti2_LU
#define  TRTI2_LN qtrti2_LN
#define  TRTRI_UU_SINGLE qtrtri_UU_single
#define  TRTRI_UN_SINGLE qtrtri_UN_single
#define  TRTRI_LU_SINGLE qtrtri_LU_single
#define  TRTRI_LN_SINGLE qtrtri_LN_single
#define  TRTRI_UU_PARALLEL qtrtri_UU_parallel
#define  TRTRI_UN_PARALLEL qtrtri_UN_parallel
#define  TRTRI_LU_PARALLEL qtrtri_LU_parallel
#define  TRTRI_LN_PARALLEL qtrtri_LN_parallel
#elif defined(DOUBLE)
#define  POTF2_U  dpotf2_U
#define  POTF2_L  dpotf2_L
#define  LAUU2_U  dlauu2_U
#define  LAUU2_L  dlauu2_L
#define  POTRF_U_SINGLE dpotrf_U_single
#define  POTRF_L_SINGLE dpotrf_L_single
#define  POTRF_U_PARALLEL dpotrf_U_parallel
#define  POTRF_L_PARALLEL dpotrf_L_parallel
#define  LAUUM_U_SINGLE dlauum_U_single
#define  LAUUM_L_SINGLE dlauum_L_single
#define  LAUUM_U_PARALLEL dlauum_U_parallel
#define  LAUUM_L_PARALLEL dlauum_L_parallel
#define  TRTI2_UU dtrti2_UU
#define  TRTI2_UN dtrti2_UN
#define  TRTI2_LU dtrti2_LU
#define  TRTI2_LN dtrti2_LN
#define  TRTRI_UU_SINGLE dtrtri_UU_single
#define  TRTRI_UN_SINGLE dtrtri_UN_single
#define  TRTRI_LU_SINGLE dtrtri_LU_single
#define  TRTRI_LN_SINGLE dtrtri_LN_single
#define  TRTRI_UU_PARALLEL dtrtri_UU_parallel
#define  TRTRI_UN_PARALLEL dtrtri_UN_parallel
#define  TRTRI_LU_PARALLEL dtrtri_LU_parallel
#define  TRTRI_LN_PARALLEL dtrtri_LN_parallel
#else
#define  POTF2_U  spotf2_U
#define  POTF2_L  spotf2_L
#define  LAUU2_U  slauu2_U
#define  LAUU2_L  slauu2_L
#define  POTRF_U_SINGLE spotrf_U_single
#define  POTRF_L_SINGLE spotrf_L_single
#define  POTRF_U_PARALLEL spotrf_U_parallel
#define  POTRF_L_PARALLEL spotrf_L_parallel
#define  LAUUM_U_SINGLE slauum_U_single
#define  LAUUM_L_SINGLE slauum_L_single
#define  LAUUM_U_PARALLEL slauum_U_parallel
#define  LAUUM_L_PARALLEL slauum_L_parallel
#define  TRTI2_UU strti2_UU
#define  TRTI2_UN strti2_UN
#define  TRTI2_LU strti2_LU
#define  TRTI2_LN strti2_LN
#define  TRTRI_UU_SINGLE strtri_UU_single
#define  TRTRI_UN_SINGLE strtri_UN_single
#define  TRTRI_LU_SINGLE strtri_LU_single
#define  TRTRI_LN_SINGLE strtri_LN_single
#define  TRTRI_UU_PARALLEL strtri_UU_parallel
#define  TRTRI_UN_PARALLEL strtri_UN_parallel
#define  TRTRI_LU_PARALLEL strtri_LU_parallel
#define  TRTRI_LN_PARALLEL strtri_LN_parallel
#endif
#else
#ifdef XDOUBLE
#define  POTF2_U  xpotf2_U
#define  POTF2_L  xpotf2_L
#define  LAUU2_U  xlauu2_U
#define  LAUU2_L  xlauu2_L
#define  POTRF_U_SINGLE xpotrf_U_single
#define  POTRF_L_SINGLE xpotrf_L_single
#define  POTRF_U_PARALLEL xpotrf_U_parallel
#define  POTRF_L_PARALLEL xpotrf_L_parallel
#define  LAUUM_U_SINGLE xlauum_U_single
#define  LAUUM_L_SINGLE xlauum_L_single
#define  LAUUM_U_PARALLEL xlauum_U_parallel
#define  LAUUM_L_PARALLEL xlauum_L_parallel
#define  TRTI2_UU xtrti2_UU
#define  TRTI2_UN xtrti2_UN
#define  TRTI2_LU xtrti2_LU
#define  TRTI2_LN xtrti2_LN
#define  TRTRI_UU_SINGLE xtrtri_UU_single
#define  TRTRI_UN_SINGLE xtrtri_UN_single
#define  TRTRI_LU_SINGLE xtrtri_LU_single
#define  TRTRI_LN_SINGLE xtrtri_LN_single
#define  TRTRI_UU_PARALLEL xtrtri_UU_parallel
#define  TRTRI_UN_PARALLEL xtrtri_UN_parallel
#define  TRTRI_LU_PARALLEL xtrtri_LU_parallel
#define  TRTRI_LN_PARALLEL xtrtri_LN_parallel
#elif defined(DOUBLE)
#define  POTF2_U  zpotf2_U
#define  POTF2_L  zpotf2_L
#define  LAUU2_U  zlauu2_U
#define  LAUU2_L  zlauu2_L
#define  POTRF_U_SINGLE zpotrf_U_single
#define  POTRF_L_SINGLE zpotrf_L_single
#define  POTRF_U_PARALLEL zpotrf_U_parallel
#define  POTRF_L_PARALLEL zpotrf_L_parallel
#define  LAUUM_U_SINGLE zlauum_U_single
#define  LAUUM_L_SINGLE zlauum_L_single
#define  LAUUM_U_PARALLEL zlauum_U_parallel
#define  LAUUM_L_PARALLEL zlauum_L_parallel
#define  TRTI2_UU ztrti2_UU
#define  TRTI2_UN ztrti2_UN
#define  TRTI2_LU ztrti2_LU
#define  TRTI2_LN ztrti2_LN
#define  TRTRI_UU_SINGLE ztrtri_UU_single
#define  TRTRI_UN_SINGLE ztrtri_UN_single
#define  TRTRI_LU_SINGLE ztrtri_LU_single
#define  TRTRI_LN_SINGLE ztrtri_LN_single
#define  TRTRI_UU_PARALLEL ztrtri_UU_parallel
#define  TRTRI_UN_PARALLEL ztrtri_UN_parallel
#define  TRTRI_LU_PARALLEL ztrtri_LU_parallel
#define  TRTRI_LN_PARALLEL ztrtri_LN_parallel
#else
#define  POTF2_U  cpotf2_U
#define  POTF2_L  cpotf2_L
#define  LAUU2_U  clauu2_U
#define  LAUU2_L  clauu2_L
#define  POTRF_U_SINGLE cpotrf_U_single
#define  POTRF_L_SINGLE cpotrf_L_single
#define  POTRF_U_PARALLEL cpotrf_U_parallel
#define  POTRF_L_PARALLEL cpotrf_L_parallel
#define  LAUUM_U_SINGLE clauum_U_single
#define  LAUUM_L_SINGLE clauum_L_single
#define  LAUUM_U_PARALLEL clauum_U_parallel
#define  LAUUM_L_PARALLEL clauum_L_parallel
#define  TRTI2_UU ctrti2_UU
#define  TRTI2_UN ctrti2_UN
#define  TRTI2_LU ctrti2_LU
#define  TRTI2_LN ctrti2_LN
#define  TRTRI_UU_SINGLE ctrtri_UU_single
#define  TRTRI_UN_SINGLE ctrtri_UN_single
#define  TRTRI_LU_SINGLE ctrtri_LU_single
#define  TRTRI_LN_SINGLE ctrtri_LN_single
#define  TRTRI_UU_PARALLEL ctrtri_UU_parallel
#define  TRTRI_UN_PARALLEL ctrtri_UN_parallel
#define  TRTRI_LU_PARALLEL ctrtri_LU_parallel
#define  TRTRI_LN_PARALLEL ctrtri_LN_parallel
#endif
#endif

#endif
#endif
