/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include "common.h"

#ifdef XDOUBLE
#define ERROR_NAME "QGETRS"
#elif defined(DOUBLE)
#define ERROR_NAME "DGETRS"
#else
#define ERROR_NAME "SGETRS"
#endif

static blasint (*getrs_single[])(long, long, FLOAT *, long, FLOAT *, long, blasint *, FLOAT *, FLOAT *) ={
  GETRS_N_SINGLE, GETRS_T_SINGLE,
};

#ifdef SMP
static blasint (*getrs_parallel[])(long, long, FLOAT *, long, FLOAT *, long, blasint *, FLOAT *) ={
  GETRS_N_PARALLEL, GETRS_T_PARALLEL,
};
#endif

int NAME(char *TRANS, blasint *N, blasint *NRHS, FLOAT *a, blasint *LDA,
	    blasint *ipiv, FLOAT *b, blasint *LDB, blasint *Info){

  char trans_arg = *TRANS;

  blasint n = *N;
  blasint nrhs = *NRHS;
  blasint lda = *LDA;
  blasint ldb = *LDB;
  blasint info;
  int trans;
  FLOAT *buffer;
#ifdef PPC440
  extern
#endif
  FLOAT *sa, *sb;

  info = 0;

  TOUPPER(trans_arg);
  trans = -1;

  if (trans_arg == 'N') trans = 0;
  if (trans_arg == 'T') trans = 1;
  if (trans_arg == 'R') trans = 0;
  if (trans_arg == 'C') trans = 1;

  if (ldb  < MAX(1, n)) info = 8;
  if (lda  < MAX(1, n)) info = 5;
  if (nrhs < 0) info = 3;
  if (n    < 0) info = 2;
  if (trans< 0) info = 1;

  if (info != 0) {
    BLASFUNC(xerbla)(ERROR_NAME, &info, sizeof(ERROR_NAME));
    return 0;
  }
  
  *Info = info;

  if (n == 0 || nrhs == 0) return 0;

#ifdef SMP
  if (blas_cpu_number == 0) blas_get_cpu_number();
#ifdef SMP_SERVER
  if (blas_server_avail == 0) blas_thread_init();
#endif
#endif

#ifndef PPC440
  buffer = (FLOAT *)blas_memory_alloc(0);

  sa = (FLOAT *)((long)buffer + GEMM_BUFFER_A_OFFSET);
  sb = (FLOAT *)(((long)sa + ((GEMM_P * GEMM_Q * COMPSIZE * SIZE + GEMM_ALIGN) & ~GEMM_ALIGN)) + GEMM_BUFFER_B_OFFSET);
#endif

#ifdef SMP
  if (blas_cpu_number == 1) {
#endif

    (getrs_single[trans])(n, nrhs, a, lda, b, ldb, ipiv, sa, sb);

#ifdef SMP
  } else {
    (getrs_parallel[trans])(n, nrhs, a, lda, b, ldb, ipiv, buffer);
  }
#endif

#ifndef PPC440
  blas_memory_free(buffer);
#endif

  return 0;
  
}
