/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include "common.h"

static void inner_thread(long m, long n, long k, FLOAT dummy1, FLOAT dummy2,
			 FLOAT *a, long lda, blasint *ipiv, long dummy4,
			 FLOAT *b, long ldb, void *buffer){

  FLOAT *sa = (FLOAT *)((void *)buffer + GEMM_BUFFER_A_OFFSET);
  FLOAT *sb = (FLOAT *)((void *)buffer + GEMM_BUFFER_B_OFFSET
			+ ((GEMM_P * GEMM_Q * COMPSIZE * SIZE +
			    GEMM_BUFFER_A_OFFSET + GEMM_ALIGN) & ~GEMM_ALIGN));

#if   TRANS == 1
      LASWP_PLUS (n, 1, m, ZERO, ZERO, b, ldb, NULL, 0, ipiv, 1);
      TRSM_LNLU  (m, n, 0, ZERO, ZERO, a, lda, NULL, 0, b,    ldb, sa, sb);
      TRSM_LNUN  (m, n, 0, ZERO, ZERO, a, lda, NULL, 0, b,    ldb, sa, sb);
#elif TRANS == 2
      TRSM_LTUN  (m, n, 0, ZERO, ZERO, a, lda, NULL, 0, b,    ldb, sa, sb);
      TRSM_LTLU  (m, n, 0, ZERO, ZERO, a, lda, NULL, 0, b,    ldb, sa, sb);
      LASWP_MINUS(n, 1, m, ZERO, ZERO, b, ldb, NULL, 0, ipiv, -1);
#elif TRANS == 3
      LASWP_PLUS (n, 1, m, ZERO, ZERO, b, ldb, NULL, 0, ipiv, 1);
      TRSM_LRLU  (m, n, 0, ZERO, ZERO, a, lda, NULL, 0, b,    ldb, sa, sb);
      TRSM_LRUN  (m, n, 0, ZERO, ZERO, a, lda, NULL, 0, b,    ldb, sa, sb);
#else
      TRSM_LCUN  (m, n, 0, ZERO, ZERO, a, lda, NULL, 0, b,    ldb, sa, sb);
      TRSM_LCLU  (m, n, 0, ZERO, ZERO, a, lda, NULL, 0, b,    ldb, sa, sb);
      LASWP_MINUS(n, 1, m, ZERO, ZERO, b, ldb, NULL, 0, ipiv, -1);
#endif

}

blasint CNAME(long n, long nrhs, FLOAT *a, long lda, FLOAT *b, long ldb, blasint *ipiv, FLOAT *buffer){
  int mode;

#if TRANS == 1
    if (nrhs == 1){
      LASWP_PLUS(nrhs, 1,    n, ZERO, ZERO, b, ldb, NULL, 0, ipiv, 1);
      ZTRSV_NLU (n, a, lda, b, 1, buffer);
      ZTRSV_NUN (n, a, lda, b, 1, buffer);
    } else {
#ifdef XDOUBLE
      mode  =  BLAS_XDOUBLE | BLAS_COMPLEX;
#elif defined(DOUBLE)
      mode  =  BLAS_DOUBLE  | BLAS_COMPLEX;
#else
      mode  =  BLAS_SINGLE  | BLAS_COMPLEX;
#endif  

      gemm_thread_n(mode, n, nrhs, 0, ZERO, ZERO,
		    a, lda, ipiv, 0, b, ldb, 
		    (void *)inner_thread, buffer);
    }
#elif TRANS == 2
    if (nrhs == 1){
      ZTRSV_TUN  (n, a, lda, b, 1, buffer);
      ZTRSV_TLU  (n, a, lda, b, 1, buffer);
      LASWP_MINUS(nrhs, 1,    n, ZERO, ZERO, b, ldb, NULL, 0, ipiv, -1);
    } else {
#ifdef XDOUBLE
      mode  =  BLAS_XDOUBLE | BLAS_COMPLEX | (1 << BLAS_TRANSA_SHIFT);
#elif defined(DOUBLE)
      mode  =  BLAS_DOUBLE  | BLAS_COMPLEX | (1 << BLAS_TRANSA_SHIFT);
#else
      mode  =  BLAS_SINGLE  | BLAS_COMPLEX | (1 << BLAS_TRANSA_SHIFT);
#endif  

      gemm_thread_n(mode, n, nrhs, 0, ZERO, ZERO,
		  a, lda, ipiv, 0, b, ldb, 
		  (void *)inner_thread, buffer);
    }
#elif TRANS == 3
    if (nrhs == 1){
      LASWP_PLUS(nrhs, 1,    n, ZERO, ZERO, b, ldb, NULL, 0, ipiv, 1);
      ZTRSV_RLU  (n, a, lda, b, 1, buffer);
      ZTRSV_RUN  (n, a, lda, b, 1, buffer);
    } else {
#ifdef XDOUBLE
      mode  =  BLAS_XDOUBLE | BLAS_COMPLEX;
#elif defined(DOUBLE)
      mode  =  BLAS_DOUBLE  | BLAS_COMPLEX;
#else
      mode  =  BLAS_SINGLE  | BLAS_COMPLEX;
#endif  

      gemm_thread_n(mode, n, nrhs, 0, ZERO, ZERO,
		    a, lda, ipiv, 0, b, ldb, 
		    (void *)inner_thread, buffer);
    }
#else
    if (nrhs == 1){
      ZTRSV_CUN  (n, a, lda, b, 1, buffer);
      ZTRSV_CLU  (n, a, lda, b, 1, buffer);
      LASWP_MINUS(nrhs, 1,    n, ZERO, ZERO, b, ldb, NULL, 0, ipiv, -1);
    } else {
#ifdef XDOUBLE
      mode  =  BLAS_XDOUBLE | BLAS_COMPLEX | (1 << BLAS_TRANSA_SHIFT);
#elif defined(DOUBLE)
      mode  =  BLAS_DOUBLE  | BLAS_COMPLEX | (1 << BLAS_TRANSA_SHIFT);
#else
      mode  =  BLAS_SINGLE  | BLAS_COMPLEX | (1 << BLAS_TRANSA_SHIFT);
#endif  

      gemm_thread_n(mode, n, nrhs, 0, ZERO, ZERO,
		  a, lda, ipiv, 0, b, ldb, 
		  (void *)inner_thread, buffer);
    }
#endif
   return 0;
  }
