/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include "common.h"

#ifdef XDOUBLE
#define ERROR_NAME "XPOTRF"
#elif defined(DOUBLE)
#define ERROR_NAME "ZPOTRF"
#else
#define ERROR_NAME "CPOTRF"
#endif

static blasint (*potrf_single[])(long, FLOAT *, long, FLOAT *, FLOAT *) ={
  POTRF_U_SINGLE, POTRF_L_SINGLE,
};

#ifdef SMP
static blasint (*potrf_parallel[])(long, FLOAT *, long, FLOAT *) ={
  POTRF_U_PARALLEL, POTRF_L_PARALLEL,
};
#endif

int NAME(char *UPLO, blasint *N, FLOAT *a, blasint *LDA, blasint *Info){

  blasint uplo_arg = *UPLO;
  blasint n    = *N;
  blasint lda  = *LDA;
  blasint uplo;
  blasint info;
  FLOAT *buffer;
#ifdef PPC440
  extern
#endif
  FLOAT *sa, *sb;
  
  TOUPPER(uplo_arg);

  uplo = -1;
  if (uplo_arg == 'U') uplo = 0;
  if (uplo_arg == 'L') uplo = 1;

  info  = 0;
  if (lda  < MAX(1,n)) info = 4;
  if (n    < 0)        info = 2;
  if (uplo < 0)        info = 1;
  if (info) {
    BLASFUNC(xerbla)(ERROR_NAME, &info, sizeof(ERROR_NAME));
    *Info = - info;
    return 0;
  }

  *Info = 0;

  if (n == 0) return 0;
  
#ifdef SMP
  if (blas_cpu_number == 0) blas_get_cpu_number();
#ifdef SMP_SERVER
  if (blas_server_avail == 0) blas_thread_init();

#endif
#endif

#ifndef PPC440
  buffer = (FLOAT *)blas_memory_alloc(0);

  sa = (FLOAT *)((long)buffer + GEMM_BUFFER_A_OFFSET);
  sb = (FLOAT *)(((long)sa + ((GEMM_P * GEMM_Q * COMPSIZE * SIZE + GEMM_ALIGN) & ~GEMM_ALIGN)) + GEMM_BUFFER_B_OFFSET);
#endif

#ifdef SMP
  if (blas_cpu_number == 1) {
#endif    

    *Info = (potrf_single[uplo])(n, a, lda, sa, sb);

#ifdef SMP
  } else {
    *Info = (potrf_parallel[uplo])(n, a, lda, buffer);
  }
#endif

#ifndef PPC440
  blas_memory_free(buffer);
#endif

  return 0;
}
