/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include "common.h"

static FLOAT dp1 =  1.;
static FLOAT dm1 = -1.;

#ifdef ITANIUM2
#undef GEMM_Q
#define GEMM_Q GEMM_P
#endif

#ifdef UNIT
#define TRTI2	TRTI2_UU
#define TRMM	TRMM_LNUU
#define TRSM	TRSM_RNUU
#else
#define TRTI2	TRTI2_UN
#define TRMM	TRMM_LNUN
#define TRSM	TRSM_RNUN
#endif

static void inner_thread1(long m, long n, long k, FLOAT dummy1,
#ifdef COMPLEX
			 FLOAT dummy2,
#endif
			 FLOAT *a, long lda, FLOAT *dummy3, long dummy4,
			 FLOAT *b, long ldb, void *buffer){

  long is, min_i;
  FLOAT *sa, *sb;
  sa = (FLOAT *)((long)buffer + GEMM_BUFFER_A_OFFSET);
  sb = (FLOAT *)((long)buffer + GEMM_BUFFER_B_OFFSET
		 + ((GEMM_P * GEMM_Q * COMPSIZE * SIZE + GEMM_BUFFER_A_OFFSET * COMPSIZE + GEMM_ALIGN) & ~GEMM_ALIGN));
  
  TRSM_OUNCOPY(n, n, a, lda, 0, sb);
  
  for(is = 0; is < m; is += GEMM_P){
    min_i = m - is;
    if (min_i > GEMM_P) min_i = GEMM_P;
    
    NEG_TCOPY(n, min_i, b + is * COMPSIZE, ldb, sa);
    
    TRSM_KERNEL_RN(min_i, n, n, dm1,
#ifdef COMPLEX
		   ZERO,
#endif
		   sa, sb, b + is * COMPSIZE, ldb, 0);
  }
}

#define GEMM_PQ  MAX(GEMM_P, GEMM_Q)

static void inner_thread2(long m, long n, long k, FLOAT dummy1,
#ifdef COMPLEX
			 FLOAT dummy2,
#endif
			 FLOAT *a, long lda, FLOAT *b, long ldb,
			 FLOAT *c, long ldc, void *buffer){

  long is, js;
  long min_i, min_j;
  FLOAT *sa1, *sa2, *sb;

  sa1 = (FLOAT *)((long)buffer + GEMM_BUFFER_A_OFFSET);
  sa2 = (FLOAT *)((((long)sa1
		    + GEMM_PQ * GEMM_Q * COMPSIZE * SIZE + 255) & ~255)
		  + GEMM_BUFFER_A_OFFSET);
  sb  = (FLOAT *)((((long)sa2
		    + GEMM_P  * GEMM_Q * COMPSIZE * SIZE + 255) & ~255)
		  + GEMM_BUFFER_B_OFFSET);
  
  TRMM_IUTCOPY(k, k, a + m * COMPSIZE, lda, 0, 0, sa1);

  for(js = 0; js < n; js += GEMM_R){
    min_j = n - js;
    if (min_j > GEMM_R) min_j = GEMM_R;
    
    GEMM_ONCOPY(k, min_j, b + js * ldb * COMPSIZE, ldb, sb);
    
    for(is = 0; is < m; is += GEMM_P){
      min_i = m - is;
      if (min_i > GEMM_P) min_i = GEMM_P;
      
      GEMM_ITCOPY(k, min_i, a + is * COMPSIZE, lda, sa2);
      
      GEMM_KERNEL_N
	(min_i, min_j, k,
	 dp1,
#ifdef COMPLEX
	 ZERO,
#endif
	 sa2, sb, c + (is + js * ldc) * COMPSIZE, ldc
	 ); 
    }

    for(is = 0; is < k; is += GEMM_P){
      min_i = k - is;
      if (min_i > GEMM_P) min_i = GEMM_P;
      
      TRMM_KERNEL_LN(min_i, min_j, k, dp1,
#ifdef COMPLEX
		     ZERO,
#endif
		     sa1 + is * k * COMPSIZE, sb, b + is + js * ldb * COMPSIZE, ldb, is); 
    }
    
  }
}


blasint CNAME(long  n, FLOAT *a, long lda, FLOAT *buffer){

  long info;
  long bk, i, blocking;
  int mode;

#ifndef COMPLEX
#ifdef XDOUBLE
  mode  =  BLAS_XDOUBLE | BLAS_REAL;
#elif defined(DOUBLE)
  mode  =  BLAS_DOUBLE  | BLAS_REAL;
#else
  mode  =  BLAS_SINGLE  | BLAS_REAL;
#endif  
#else
#ifdef XDOUBLE
  mode  =  BLAS_XDOUBLE | BLAS_COMPLEX;
#elif defined(DOUBLE)
  mode  =  BLAS_DOUBLE  | BLAS_COMPLEX;
#else
  mode  =  BLAS_SINGLE  | BLAS_COMPLEX;
#endif  
#endif

  if (n <= 32) {
    info = TRTI2(n, a, lda, buffer);
    return info;
  }

  blocking = GEMM_Q;
  if (n <= 4 * GEMM_Q) blocking = n / 4;

  for (i = 0; i < n; i += blocking) {
    
    bk = MIN(blocking, n - i);

    if (i > 0) {
      trsm_thread( mode | BLAS_RSIDE,
		   i, bk, ZERO, ZERO,
		   a + (i + i * lda) * COMPSIZE, lda, 
		   a + (    i * lda) * COMPSIZE, lda, 
		   (void *)inner_thread1,
		   buffer);
    }

    CNAME(bk, a + (i + i * lda) * COMPSIZE, lda, buffer);
    
    gemm_thread_n( mode,
		 i, n - i - bk, bk, dp1, ZERO,
		 a + (     i       * lda) * COMPSIZE, lda,
		 a + (i + (i + bk) * lda) * COMPSIZE, lda,
		 a + (    (i + bk) * lda) * COMPSIZE, lda, 
		 (void *)inner_thread2,
		 buffer);
  }
  
  return 0;
}
