/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include "common.h"

static FLOAT dp1 =  1.;
static FLOAT dm1 = -1.;

#ifdef UNIT
#define TRTI2	TRTI2_UU
#else
#define TRTI2	TRTI2_UN
#endif

#define GEMM_PQ  MAX(GEMM_P, GEMM_Q)
#define REAL_GEMM_R (GEMM_R - 2 * GEMM_PQ)

blasint CNAME(long  n, FLOAT *a, long lda, FLOAT *sa, FLOAT *sb){

  long i, is, min_i, start_is;
  long ls, min_l;
  long bn, bk;
  long blocking;

  FLOAT *sa_trsm = (FLOAT *)((long)sb);
  FLOAT *sa_trmm = (FLOAT *)((((long)sb
			       + GEMM_PQ * GEMM_Q * COMPSIZE * SIZE + 255) & ~255)
			     + GEMM_BUFFER_A_OFFSET);
  FLOAT *sb_gemm = (FLOAT *)((((long)sa_trmm
			       + GEMM_PQ * GEMM_Q * COMPSIZE * SIZE + 255) & ~255)
			     + GEMM_BUFFER_B_OFFSET);

  if (n <= DTB_ENTRIES) {
    TRTI2(n, a, lda, sb);
    return 0;
  }

  blocking = n / 4;
  if (blocking > GEMM_Q) blocking = GEMM_Q;

  for (i = 0; i < n; i += blocking) {
    bk = MIN(blocking, n - i);
    bn = n - bk - i;
    
    if (i  > 0) TRSM_OUNCOPY(bk, bk, a + (i + i * lda) * COMPSIZE, lda, 0, sa_trsm);

    CNAME(bk, a + (i + i * lda) * COMPSIZE, lda, sa, sa_trmm);

    if (bn > 0) {
      TRMM_IUTCOPY(bk, bk, a + (i + i * lda) * COMPSIZE, lda, 0, 0, sa_trmm);

      for (ls = 0; ls < bn; ls += REAL_GEMM_R) {
	min_l = bn - ls;
	if (min_l > REAL_GEMM_R) min_l = REAL_GEMM_R;
	
	GEMM_ONCOPY (bk, min_l, a + (i + (i + bk + ls) * lda) * COMPSIZE, lda, sb_gemm);
	
	if (i > 0) {
	  for (is = 0; is < i; is += GEMM_P) {
	    min_i = i - is;
	    if (min_i > GEMM_P) min_i = GEMM_P;
	    
	  if (ls == 0) {
	    NEG_TCOPY (bk, min_i, a + (is + i * lda) * COMPSIZE, lda, sa);
	    
	    TRSM_KERNEL_RN(min_i, bk, bk, dm1, 
#ifdef COMPLEX
			   ZERO,
#endif
			   sa, sa_trsm,
			   a + (is + i * lda) * COMPSIZE, lda, 0);
	  } else {
	    GEMM_ITCOPY (bk, min_i, a + (is + i * lda) * COMPSIZE, lda, sa);
	  }
	    
	  GEMM_KERNEL_N(min_i, min_l, bk, dp1, 
#ifdef COMPLEX
			ZERO,
#endif
			sa, sb_gemm,
			a + (is + (i + bk + ls) * lda) * COMPSIZE, lda);
	  }
	}
	
	start_is = 0;
	while (start_is < bk) start_is += GEMM_P;
	start_is -= GEMM_P;

	for (is = 0; is < bk; is += GEMM_P) {
	  min_i = bk - is;
	  if (min_i > GEMM_P) min_i = GEMM_P;
	  
	  TRMM_KERNEL_LN(min_i, min_l, bk, dp1, 
#ifdef COMPLEX
			 ZERO,
#endif
			 sa_trmm + is * bk * COMPSIZE, sb_gemm,
			 a + (i + is + (i + bk + ls) * lda) * COMPSIZE, lda, is);
	}
      }
    } else {
      if (i > 0) {
	for (is = 0; is < i; is += GEMM_P) {
	  min_i = i - is;
	  if (min_i > GEMM_P) min_i = GEMM_P;
	  
	  NEG_TCOPY (bk, min_i, a + (is + i * lda) * COMPSIZE, lda, sa);
	  
	  TRSM_KERNEL_RN(min_i, bk, bk, dm1, 
#ifdef COMPLEX
			 ZERO,
#endif
			 sa, sa_trsm,
			 a + (is + i * lda) * COMPSIZE, lda, 0);
	}
      }
    }
  }

  return 0;
}
