/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include <ctype.h>
#include "common.h"

#ifndef COMPLEX
#ifdef XDOUBLE
#define SYMM_IUTCOPY qsymm_iutcopy
#define SYMM_ILTCOPY qsymm_iltcopy
#define SYMM_OUTCOPY qsymm_outcopy
#define SYMM_OLTCOPY qsymm_oltcopy
#elif defined(DOUBLE)
#define SYMM_IUTCOPY dsymm_iutcopy
#define SYMM_ILTCOPY dsymm_iltcopy
#define SYMM_OUTCOPY dsymm_outcopy
#define SYMM_OLTCOPY dsymm_oltcopy
#else
#define SYMM_IUTCOPY ssymm_iutcopy
#define SYMM_ILTCOPY ssymm_iltcopy
#define SYMM_OUTCOPY ssymm_outcopy
#define SYMM_OLTCOPY ssymm_oltcopy
#endif
#else
#ifdef XDOUBLE
#define SYMM_IUTCOPY xsymm_iutcopy
#define SYMM_ILTCOPY xsymm_iltcopy
#define SYMM_OUTCOPY xsymm_outcopy
#define SYMM_OLTCOPY xsymm_oltcopy
#elif defined(DOUBLE)
#define SYMM_IUTCOPY zsymm_iutcopy
#define SYMM_ILTCOPY zsymm_iltcopy
#define SYMM_OUTCOPY zsymm_outcopy
#define SYMM_OLTCOPY zsymm_oltcopy
#else
#define SYMM_IUTCOPY csymm_iutcopy
#define SYMM_ILTCOPY csymm_iltcopy
#define SYMM_OUTCOPY csymm_outcopy
#define SYMM_OLTCOPY csymm_oltcopy
#endif
#endif

#if GEMM_UNROLL_M == GEMM_UNROLL_N
#undef  SYMM_IUTCOPY
#undef  SYMM_ILTCOPY
#define SYMM_IUTCOPY SYMM_OUTCOPY
#define SYMM_ILTCOPY SYMM_OLTCOPY
#endif

#ifndef SWITCHING
#define SWITCHING	(GEMM_P *  2)
#endif

#define RRR  GEMM_UNROLL_N

#ifndef RSIDE
#define A	a
#define LDA	lda
#define B	b
#define LDB	ldb
#define K	m
#else
#define A	b
#define LDA	ldb
#define B	a
#define LDB	lda
#define K	n
#endif

int CNAME(long m, long n, long dummy1, FLOAT alpha_r, 
#ifdef COMPLEX
		  FLOAT alpha_i,
#endif
	  FLOAT *a, long lda, FLOAT *b, long ldb, FLOAT *c, long ldc, FLOAT *sa, FLOAT *sb){

  long ls, is, js;
  long min_l, min_i, min_j;
  FLOAT *real_sa, *real_sb;

#if defined(SWITCHING)
  if (m >= SWITCHING) {
#endif
  for(ls = 0; ls < K; ls += min_l){
    min_l = K - ls;
    if (min_l >= GEMM_Q * 2) {
      min_l = GEMM_Q;
    } else 
      if (min_l > GEMM_Q) {
#ifndef RSIDE
	min_l = (min_l / 2 + GEMM_UNROLL_M - 1) & ~(GEMM_UNROLL_M - 1);
#else
	min_l = (min_l / 2 + GEMM_UNROLL_N - 1) & ~(GEMM_UNROLL_N - 1);
#endif
      }
    
    for(js = 0; js < n; js += GEMM_R){
      min_j = n - js;
      if (min_j > GEMM_R) min_j = GEMM_R;
      
      if (B != NULL) {
#ifndef RSIDE
	GEMM_ONCOPY(min_l, min_j, B + (ls + js * LDB) * COMPSIZE, LDB, sb);
#else
#ifndef LOWER
	  SYMM_OUTCOPY(min_l, min_j, B, LDB, ls, js, sb);
#else
	  SYMM_OLTCOPY(min_l, min_j, B, LDB, ls, js, sb);
#endif
#endif
	real_sb = sb;
      } else {
	real_sb = sb + ls * js * COMPSIZE;
      }
      
      for(is = 0; is < m; is += min_i){
	min_i = m - is;
	if (min_i >= GEMM_P * 2) {
	  min_i = GEMM_P;
	} else 
	  if (min_i > GEMM_P) {
	    min_i = (min_i / 2 + GEMM_UNROLL_M - 1) & ~(GEMM_UNROLL_M - 1);
	  }

	if (a != NULL) {
#ifndef RSIDE
#ifndef LOWER
	  SYMM_IUTCOPY(min_l, min_i, A, LDA, ls, is, sa);
#else
	  SYMM_ILTCOPY(min_l, min_i, A, LDA, ls, is, sa);
#endif
#else
	  GEMM_ITCOPY(min_l, min_i, A + (is + ls * LDA) * COMPSIZE, LDA, sa);
#endif
	  real_sa = sa;
	} else {
	  real_sa = sa + ls * is * COMPSIZE;
	}

#if defined(NN) || defined(NT) || defined(TN) || defined(TT)
	GEMM_KERNEL_N
#endif
#if defined(CN) || defined(CT) || defined(RN) || defined(RT)
        GEMM_KERNEL_L
#endif
#if defined(NC) || defined(TC) || defined(NR) || defined(TR)
        GEMM_KERNEL_R
#endif
#if defined(CC) || defined(CR) || defined(RC) || defined(RR)
        GEMM_KERNEL_B
#endif
	  (min_i, min_j, min_l, alpha_r, 
#ifdef COMPLEX
		      alpha_i,
#endif
	   real_sa, real_sb, c + (is + js * ldc) * COMPSIZE, ldc); 
	
      } /* end of ls */
    } /* end of is */
  } /* end of js */
#if defined(SWITCHING)
  } else {
    for(ls = 0; ls < K; ls += min_l){
      min_l = K - ls;
    if (min_l >= GEMM_Q * 2) {
      min_l = GEMM_Q;
    } else 
      if (min_l > GEMM_Q) {
#ifndef RSIDE
	min_l = (min_l / 2 + GEMM_UNROLL_M - 1) & ~(GEMM_UNROLL_M - 1);
#else
	min_l = (min_l / 2 + GEMM_UNROLL_N - 1) & ~(GEMM_UNROLL_N - 1);
#endif
      }
      
      for(is = 0; is < m; is += min_i){
	min_i = m - is;
	if (min_i >= GEMM_P * 2) {
	  min_i = GEMM_P;
	} else 
	  if (min_i > GEMM_P) {
	    min_i = (min_i / 2 + GEMM_UNROLL_M - 1) & ~(GEMM_UNROLL_M - 1);
	  }

	if (A != NULL) {
#ifndef RSIDE
#ifndef LOWER
	  SYMM_IUTCOPY(min_l, min_i, A, LDA, ls, is, sa);
#else
	  SYMM_ILTCOPY(min_l, min_i, A, LDA, ls, is, sa);
#endif
#else
	  GEMM_ITCOPY(min_l, min_i, b + (is + ls * ldb) * COMPSIZE, ldb, sa);
#endif
	  real_sa = sa;
	} else {
	  real_sa = sa + ls * is * COMPSIZE;
	}

      for(js = 0; js < n; js += min_j){
	min_j = n - js;
	if (min_j >= RRR * 2) {
	  min_j = RRR;
	} else 
	  if (min_j > RRR) {
	    min_j = (min_j / 2 + GEMM_UNROLL_N - 1) & ~(GEMM_UNROLL_N - 1);
	  }
	
	if (B != NULL) {
#ifndef RSIDE
	  GEMM_ONCOPY(min_l, min_j, B + (ls + js * LDB) * COMPSIZE, LDB, sb);
#else
#ifndef LOWER
	    SYMM_OUTCOPY(min_l, min_j, a, lda, ls, js, sb);
#else
	    SYMM_OLTCOPY(min_l, min_j, a, lda, ls, js, sb);
#endif
#endif
	  real_sb = sb;
	} else {
	  real_sb = sb + ls * js * COMPSIZE;
	}
	
#if defined(NN) || defined(NT) || defined(TN) || defined(TT)
	GEMM_KERNEL_N
#endif
#if defined(CN) || defined(CT) || defined(RN) || defined(RT)
	  GEMM_KERNEL_L
#endif
#if defined(NC) || defined(TC) || defined(NR) || defined(TR)
	  GEMM_KERNEL_R
#endif
#if defined(CC) || defined(CR) || defined(RC) || defined(RR)
	  GEMM_KERNEL_B
#endif
	  (min_i, min_j, min_l, alpha_r, 
#ifdef COMPLEX
	   alpha_i,
#endif
	   real_sa, real_sb, c + (is + js * ldc) * COMPSIZE, ldc); 
	
      } /* end of ls */
      } /* end of is */
    } /* end of js */
  }
#endif

  return 0;
}
